'''
02-February-2023
Bradán Lane STUDIO
2023 eChallengeCoin Story Challenge Knight's Tour Solver (just an aid, really)


This python program is quite a hack of bits and pieces from various sources.

The original algorithm: https://rosettacode.org/wiki/Knight%27s_tour#Python


This code was first modified to maintain a list of possible branches and then
backtrack to try each one. This is implemented with recursion and deep copy
which is very memory intensive (but I said this was a hack).

While the algorithm is fast, each iteration is a depth-first solution.
Since we know the solution results in text words, we ideally want a 
breadth-first solver. This turned out to be too much work.


The story clue has been formatted so every work is capitalized.
We take advantage of this fact to solved for words rather than solve for
the whole clue.

The solver starts with upper left 0,0 (A8 in chess terms).
It has two nested loops to scan X and Y looking for capital letters.
When it find a capitol letter, it runs a solver of up to N letters,
looking for the next capital letter.

The list of all possible letter sequences between two capital letters
was another horrible mess so the solver applies a simple English
language dictionary test. If the word is found, then it is saved,
otherwise it is discarded.

This mean the clue had to be tested for valid English (US) language
words. The clue was gradually tweaked so it passed this test.

The solver repeats the process, scanning the entire 8x8 grid.


Finally, the python program was extended to allow it to take 
a starting X,Y and an optional word length and solve just one segment.
The X and Y are zero-based values.
With this use-case, a user can run for the first word, analyze
the results, and then run for the logical next word / start position.

Like I said, this is a hack of bits and pieces ... but it does find
the words in the clue and enough of them are unique that it is possible
to solve the Knight's Tour and move forward.
'''



import sys
import copy
from datetime import datetime
import re
import enchant

# code from the 2023 eChallengeCoin story challenge #3
code = [
    [	'V',	'n',	'x',	'F',	's',	'J',	'a',	'i',		],	
	[	'y',	'd',	'i',	'u',	'B',	'r',	'F',	't',		],	
	[	'A',	'.',	'O',	'o',	'i',	'i',	'e',	'y',		],	
	[	'l',	'B',	's',	'd',	'i',	'd',	'r',	'r',		],	
	[	'e',	'e',	'n',	'T',	'F',	'n',	'A',	'n',		],	
	[	'e',	'u',	'a',	't',	'S',	'a',	'o',	'a',		],	
	[	'r',	'm',	'Y',	't',	'r',	'H',	'a',	'n',		],	
	[	'o',	'A',	'A',	'o',	'y',	'd',	'm',	'm',		],	]



solutions = []
english_words = enchant.Dict("en_US")

_kmoves = ((2,1), (1,2), (-1,2), (-2,1), (-2,-1), (-1,-2), (1,-2), (2,-1)) 

def chess2index(chess, boardsize):
    'Convert Algebraic chess notation to internal index format'
    chess = chess.strip().lower()
    x = ord(chess[0]) - ord('a')
    y = boardsize - int(chess[1:])
    return (x, y)

def outputdecodestring(board, moves, boardsize):
    global solutions

    sorted_board = dict(sorted(board.items(), key=lambda item: item[1]))

    tail = dict(list(sorted_board.items())[-(moves):])
    #print(tail)

    msg = ""
    for entry in tail:
        x = entry[0]
        y = entry[1]
        msg = msg + code[x][y]
    '''
        the msg is the whole board
        we only want the last 'moves' characters
        we then want to see if there is a capitalized letter (or period) *after* the first character
    '''
    #print(msg)
    part = re.search('[A-Z\.]', msg[1:])
    if part is None:
        return

    msg_word = msg[:part.end()]
    if len(msg_word) == 0:
        return

    if not english_words.check(msg_word):
        return

    pos = list(tail)[part.end()]
    next_letter = code[pos[0]][pos[1]]
    '''
    if (msg_word == 'xyzzy'): # you can add a word here to verify it is found
        print ('%s: next (%d,%d)' % (msg_word, pos[1], pos[0])
    '''
    msg_word += '(%d,%d %s)' % (pos[1], pos[0], next_letter)

    if not msg_word in solutions:
        solutions.append(msg_word)
        #print(msg_word)

    '''
    now = datetime.now()
    now_text = now.strftime("%H:%M:%S")
    answer = "%s: %s" % (now_text, msg)
    print(answer)

    f = open('toursolver.txt', 'a')
    f.write(answer)
    f.write("\n")
    f.close()
    '''

def boardstring(board, boardsize):
    r = range(boardsize)
    lines = ''
    for y in r:
        lines += '\n' + ','.join('%2i' % board[(x,y)] if board[(x,y)] else '  ' for x in r)
    return lines
    
def knightmoves(board, P, boardsize, level):
    Px, Py = P
    kmoves = set((Px+x, Py+y) for x,y in _kmoves)
    kmoves = set( (x,y)
                  for x,y in kmoves
                  if 0 <= x < boardsize
                     and 0 <= y < boardsize
                     and not board[(x,y)] )
    #print (level, kmoves)
    return kmoves

def accessibility(board, P, boardsize):
    access = []
    brd = copy.deepcopy(board)
    for pos in knightmoves(board, P, boardsize, 0):
        brd[pos] = -1
        access.append( (len(knightmoves(brd, pos, boardsize, 1)), pos) )
        brd[pos] = 0
    #print (access)
    #print()
    return access
    
def sub_tour(move, moves, board, P, boardsize):
    while move <= moves:
        answers = accessibility(board, P, boardsize)  
        sorted_answers = sorted(answers)
        size = len(sorted_answers)
        index = 0
        success = False
        for answer in sorted_answers:
            index += 1
            #print ("answer (%2d of %2d): " % (index, size), answer)
            P = answer[1]
            brd = copy.deepcopy(board)
            brd[P] = move
            success = sub_tour(move+1, moves, brd, P, boardsize)
            #input("\n%2d next: " % (move))
        return False # didn't find an answer
    #print("done: ", board)
    outputdecodestring(board, moves, boardsize)
    #input("\n%2d again: " % (move))
    return True # found an answer
    #print(boardstring(board, boardsize))


def knights_tour(start, moves, boardsize, _debug=False):
    board = {(x,y):0 for x in range(boardsize) for y in range(boardsize)}
    move = 1
    #P = chess2index(start, boardsize)
    P = start
    board[P] = move
    move += 1
    if _debug:
        print(boardstring(board, boardsize))
    sub_tour(move, moves, board, P, boardsize)
    return


if __name__ == '__main__':
    if 0:
        # this section is for debugging
        # assign a string to 'clue' to test if the words are in the dictionary
        clue = ""
        clue_words = clue.split(' ')
        for word in clue_words:
            print("%s is %s" % (word, ("good" if english_words.check(word) else "bad")))
    else:
        # this section is the actual solver / helper program
        boardsize = 8
        msg = ""
        moves = 9
        start = (7,0)

        if len(sys.argv) > 2:
            # run the solver for a single word
            y = int(sys.argv[1])
            x = int(sys.argv[2])
            start = (x,y)
            if len(sys.argv) > 3:
                moves = int(sys.argv[3])
            print("Knight's Tour Solver: Single Word starting with %s at (%d,%d) for %d letters" % (code[x][y], y, x, moves))
            knights_tour(start, moves, boardsize, _debug=False)
            solutions.sort()
            pos_xy = start[::-1]
            print (str(pos_xy), ': ', solutions)
        else:
            # run the solver for every word in the clue
            x = 0
            y = 0
            print("Knight's Tour Solver: All Words starting with %s at (%d,%d) for %d letters" % (code[x][y], y, x, moves))
            for y in range(boardsize):
                for x in range(boardsize):
                    code_letter = code[x][y]
                    if code_letter.isupper():
                        start = (x, y)
                        #print('Tour Solver starting with %s at %s for %d letters' % (code[x][y], str(start), moves))
                        knights_tour(start, moves, boardsize, _debug=False)
                        solutions.sort()
                        pos_xy = start[::-1]
                        print (str(pos_xy), ': ', solutions)
                        #print()
                        solutions = []
